<?php
/**
 * Copyright (c) 2005-2008 GForge, LLC
 *
 * @version   $Id:
 * @author Federico Tarantini federico@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

//require_once('SystemUtils.php');
require_once('plugins/scmgit/lib/exceptions/GITException.php');
require_once('plugins/scm/lib/SCMRepository.php');


class GitRepository extends SCMRepository {
	private $repositoryPath;
	private $gitRoot;
	public $hostname;
	private $revisionCache = array();

	public function __construct(Project $project) {
		parent::__construct($project);

		$this->gitRoot = GFConf::get('scmgit.localGitRoot');
		$this->repositoryPath = SystemUtils::buildFilePath($this->gitRoot, '/'.$this->repositoryName);
		$this->gitpath = GFConf::get('scmgit.gitpath');
		$this->useSSH = GFConf::get('scmgit.useSSH');
	}


	/**
	 * Return the type of this repository (ie, 'cvs', 'svn', etc)
	 * @return	string
	 */
	public function getType(){
		return 'git';
	}
	
	/**
	 * Returns the for this repository
	 * @return	string
	 */
	public function getRoot(){
		return $this->repositoryPath;
	}
	
	/**
	 * Returns the root for ALL repositories
	 * @return	string
	 */
	public function getScmRoot(){
	}
	
	/**
	 * @return	bool	True if the repository is created
	 */
	public function isCreated() {
		return is_dir($this->repositoryPath);
	}
	
	public function create() {
		if ($this->isCreated()) {
			throw new GitException('Tried to create an existing repository');
		}

		if ($this->useSSH) {
			$output = array();
			$unix_group_name = $this->project->getUnixName();
			if (!SystemUtils::groupExists($unix_group_name)) {
				throw new GitException('Tried to create repository for project '.$unix_group_name.' but UNIX group doesn\'t exist');
			}
			
			mkdir($this->repositoryPath, 0770, true);

	 		$commands = array(
	 			// Sequence of commands to execute in order to create the repository
	 			'cd '.$this->repositoryPath." && ".$this->gitpath.'/git --bare init --shared',
	 			'echo "'.$this->project->getProjectName().'" > '.$this->repositoryPath.'/description'
	 		);
	 		
	 		foreach ($commands as $command) {
	 			if (SystemUtils::execute($command, $output) != 0) {
	 				throw new GitException('Error creating Git repository: '.implode("\n", $output));
	 			}
	 		}

	 		$this->updateRepository();
	 	} else {
	 		throw new GitException('Git access via DAV currently unsupported');
	 	}
	}
	
	public function updateRepository() {
		if ($this->useSSH) {
			// Update the permissions
			$unix_group_name = $this->project->getUnixName();
			
			// Update the permissions
			$unix_group_name = $this->project->getUnixName();

			SystemUtils::execute('chgrp -R '.$unix_group_name.' '.$this->repositoryPath, $output);
			SystemUtils::execute('chown nobody:'.$unix_group_name.' '.$this->repositoryPath, $output);

			$project_scm = $this->getScmConfiguration();

			if ($this->project->getIsPublic() && $project_scm->getEnableAnonRead()) {
				$this->enableAnonAccess();
			} else {
				$this->disableAnonAccess();
			}
			
			$this->updatePreReceive();
			$this->updatePostReceive();
		}
	}
	
	private function updatePreReceive() {
		// GForge manages the pre-receive script
		$unix_group_name = $this->project->getUnixName();
		$preReceiveFile = SystemUtils::buildFilePath($this->repositoryPath, '/hooks/pre-receive');

		$preReceiveCheckScript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmgit/bin/pre_receive_checks.php');

		$lines = array();
		$lines[] = '#!/bin/sh';
		$lines[] = '#BEGIN GForge pre-receive checks';
		$lines[] = 'php -q -d include_path='.ini_get('include_path').' '.$preReceiveCheckScript;
		$lines[] = '#END GForge pre-receive checks';
		
		$this->insertCommitFilters($lines);

 		$this->writeFile($preReceiveFile, implode("\n", $lines));
 		SystemUtils::execute('chmod +x '.$preReceiveFile, $output);
		SystemUtils::execute('chmod 770 '.$preReceiveFile, $output);
		chown($preReceiveFile, 'nobody');
		chgrp($preReceiveFile, $unix_group_name);
	}
	
	private function updatePostReceive() {
		// GForge manages the post-receive script
		$unix_group_name = $this->project->getUnixName();
		$postReceiveFile = SystemUtils::buildFilePath($this->repositoryPath, '/hooks/post-receive');

		$postReceiveScript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmgit/bin/post_receive.php');

		$lines = array();
		$lines[] = '#!/bin/sh';
		$lines[] = '#BEGIN GForge post-receive hooks';
		$lines[] = 'php -q -d include_path='.ini_get('include_path').' '.$postReceiveScript;
		$lines[] = '#END GForge post-receive hooks';
		
		$this->updateMail($lines);

 		$this->writeFile($postReceiveFile, implode("\n", $lines));
 		SystemUtils::execute('chmod +x '.$postReceiveFile, $output);
		SystemUtils::execute('chmod 770 '.$postReceiveFile, $output);
		chown($postReceiveFile, 'nobody');
		chgrp($postReceiveFile, $unix_group_name);
	}
	
	private function enableAnonAccess() {
	}
	
	private function disableAnonAccess() {
	}

	/**
	 * Writes to a file with locking
	 * @param	string $filePath
	 * @param	mixed $content	Contents of the file. Can be a string or an array with the lines
	 * @param	int	0 write to the file, 1 append to the file, 2 truncate, rewind and rewrite
	 */
	private function writeFile($filePath, $content, $append=0) {
		if (is_array($content)) {
			$content = implode("\n", $content);
		}
		if ($append==2) {
			$file = fopen($filePath, 'a');
			flock($file, LOCK_EX);
			ftruncate($file, 0);
			rewind($file);
			if(!empty($content)) {
				fwrite($file, $content);
			}
			flock($file, LOCK_UN);
			fclose($file);	
		} else {
			if ($append == 1) {
				$file = fopen($filePath, 'a');
			} else {
				$file = fopen($filePath, 'w');
			}
			flock($file, LOCK_EX);
			if(!empty($content)) {
				fwrite($file, $content);
			}
			flock($file, LOCK_UN);
			fclose($file);
		}
	}
	
	/**
	 * Inserts the commit filter commands
	 */
	private function insertCommitFilters(&$lines_src) {
		$lines_dest = array();

		for ($i = 0; $i < count($lines_src); $i++) {
			$line = $lines_src[$i];
			if (strpos($line, '#BEGIN GForge filters') !== false) {
				// Seek to the end of the marker
				for (; $i < count($lines_src); $i++) {
					$line = $lines_src[$i];
					if (strpos($line, '#END GForge filters') !== false) {
						break;
					}
				}
			} else {
				$lines_dest[] = $line;
			}
		}
		
		// Now append the filters

		$c = new Criteria();
		$c->add(ProjectPluginPeer::PROJECT_ID, $this->project->getProjectId());
		$c->addJoin(ProjectPluginPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->add(PluginPeer::PLUGIN_NAME, ScmgitPlugin::getName());		// Make sure we're not including filters we shouldn't
		$c->addJoin(ScmCommitFilterProjectPeer::PROJECT_ID, ProjectPluginPeer::PROJECT_ID);
		$c->addJoin(ScmCommitFilterPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->addJoin(ScmCommitFilterProjectPeer::SCM_COMMIT_FILTER_ID, ScmCommitFilterPeer::SCM_COMMIT_FILTER_ID);
		$filters = ScmCommitFilterPeer::doSelect($c);
		
		$lines_dest[] = '#BEGIN GForge filters';
		foreach ($filters as $filter) {
			$command = $filter->getFilterString();
			$command = StringUtils::unHtmlSpecialChars($command);
			$command = preg_replace('/\\$PROJECT/i', $this->project->getUnixName(), $command);
			$command = preg_replace('/\\$PATH/i', $this->getroot(), $command);
			//$command = preg_replace('/\\$FILE/i', '\\$2', $command);
			$lines_dest[] = $command;
		}
		$lines_dest[] = '#END GForge filters';
		
		$lines_src = $lines_dest;
	}
	
	public function updateMail(&$lines) {
		// Read the preferences
		$preferences = $this->getScmPreferences();

 		if (!is_null($preferences) && $preferences->getMailCommits()) {
			$this->addMail($lines);
		} else {
			//$this->removeSVNMail($lines);
		}
	}
	
	public function addMail(&$lines) {
		// First check if we have mailman configured
		if (!$this->project->usesPlugin('mailman')) {
			return;
		}
		
		// Set up the mailing lists on the Git repository
 		$commands = array(
 			// Sequence of commands to execute
 			'cd '.$this->repositoryPath." && ".$this->gitpath.'/git config hooks.mailinglist "'.$this->repositoryName.'-commits@'.GFConf::get('mailman.listsHost').'"',
 			'cd '.$this->repositoryPath." && ".$this->gitpath.'/git config hooks.emailprefix "[GForge-Git]"',
 		);
 		
 		foreach ($commands as $command) {
 			if (SystemUtils::execute($command, $output) != 0) {
 				throw new GitException('Error creating Git repository: '.implode("\n", $output));
 			}
 		}
 		
 		// Note that the actual email is sent by the post_recieve.php script
	}
	
	
	public function doLocalCheckout($destdir, $module) {
		// $module here is the branch name
		$path = $this->repositoryPath;
		if (!$module) {
			$module = 'master';			// The default module
		}
		
		if (is_dir($destdir)) {
			rmdir($destdir);			// Directory must not exist
		}
		
		$output = array();
		$cmd = $this->gitpath.'/git clone -lsn '.$this->repositoryPath.' '.$destdir;
		$error = SystemUtils::execute($cmd, $output);
		if ($error) {
			throw new Exception("Unable to checkout the code - Error code $error - Command was '$cmd'");
		}
		
		chdir($destdir);

		$cmd = $this->gitpath."/git checkout -b $module origin/$module";
		$error = SystemUtils::execute($cmd, $output);
		if ($error) {
			throw new Exception("Unable to checkout the code - Error code $error - Command was '$cmd'");
		}
	}
	
	public function getCheckedoutFileRevision($checkout_dir, $file) {
		if (!isset($this->revisionCache[$checkout_dir])) {
			chdir($checkout_dir);
			
			$output = array();
			$cmd = $this->gitpath.'/git rev-list --max-count=1 HEAD';
			$error = SystemUtils::execute($cmd, $output);
			if ($error) {
				return -1;
			}
			
			$revision = trim($output[0]);
			$this->revisionCache[$checkout_dir] = $revision;
		}
		
		return $this->revisionCache[$checkout_dir];
	}
}

?>
